program bigbuf(afile, bigbufp, bfile, output);
(* bigbuf: show last n characters of a file

  Dr. Thomas D. Schneider
  toms@alum.mit.edu
  permanent email: toms@alum.mit.edu (use only if first address fails)
  https://alum.mit.edu/www/toms/

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.01; (* of bigbuf.p 2010 Feb 05
2010 Feb 05, 1.01: finish up documentation
2010 Feb 05, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.bigbuf *)
(*
name
   bigbuf: show last n characters of a file

synopsis
   bigbuf(afile: in, bigbufp: in, bfile: out, output: out)

files

   afile:  input file to examine

   bfile:  the last buffersize characters of afile

   bigbufp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This
      allows the user to be warned if an old parameter file is used.

      buffersize (integer): number of characters to to store in a
      circular buffer

      bufferstart (integer): number of characters to display from the
      start of the buffer

   output: messages to the user

description

   Read characters into a circular buffer so that only the last
   buffersize characters are stored.  At the end of the file, show the
   first bufferstart characters of the buffer.

examples

strawberry 76% b
This is BibTeX, Version 0.99c (Web2C 7.5.5)
The top-level auxiliary file: paper.aux
The style file: unsrt.bst
Database file #1: buildrefs/troff.bib
Your field is more than 5000 characters---line 685 of file buildrefs/troff.bib
 : title = "{I protein: bacteriophage T7-coded inhibito
 :                                                     r of Escherichia
I'm skipping whatever remains of this entry
Your field is more than 5000 characters---line 964 of file buildrefs/troff.bib
 : y
 :  ear = "1982"}
I'm skipping whatever remains of this entry
Your field is more than 5000 characters---line 1198 of file buildrefs/troff.bib
 : %% McAllister, W.T., Morris, C., Rosenberg, A.H. and Studier, F.W. (1981)
 :                                                                          .  \fI
J. Mol. Biol.\fP \fB153\fP, 527-544.
I'm skipping whatever remains of this entry
Your field is more than 5000 characters---line 1500 of file buildrefs/troff.bib
 : %% Pingoud, A.
 :                (1985).  \fIEur. J. Biochem.\fP \fB147\fP, 105-109.
I'm skipping whatever remains of this entry
Illegal end of database file---line 2153 of file buildrefs/troff.bib
 : year = "1981"}
 :               
I'm skipping whatever remains of this entry

How to locate this?  One must count back 5000 characters from the end
of the file.

documentation

see also

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.bigbuf *)

var
   afile, (* file used by this program *)
   bigbufp, (* file used by this program *)
   bfile: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module bigbuf.themain *)
procedure themain(var afile, bigbufp, bfile: text);
(* the main procedure of the program *)
const
   buffermax = 10000; (* maximum buffer allowed *)
var
   parameterversion: real; (* parameter version number *)
   buffer: array [1..buffermax] of char; (* buffer size to report on *)
   buffersize: integer; (* buffer size to report on *)
   bufferstart: integer; (* amount of buffer start to report *)
   c: integer; (* character count *)
   ch: char; (* character read in *)
   l: integer; (* line count *)
   s: integer; (* store point in the buffer *)
   
begin

   writeln(output,'bigbuf ',version:4:2);
   reset(bigbufp);
   readln(bigbufp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   readln(bigbufp, buffersize);
   if buffersize < 2 then begin
      writeln(output,'minimum buffer allowed is 1');
      halt;
   end;
   if buffersize > buffermax then begin
      writeln(output,'maximum buffer allowed is ',buffermax:1);
      halt;
   end;

   readln(bigbufp, bufferstart);
   if bufferstart < 2 then begin
      writeln(output,'minimum buffer start allowed is 1');
      halt;
   end;
   if bufferstart > buffermax then begin
      writeln(output,'maximum buffer start allowed is ',buffermax:1);
      halt;
   end;

   reset(afile); 
   c := 0;
   l := 0;
   s := 0;
   while not eof (afile) do begin
      if eoln(afile)
      then begin
         readln(afile);
         l := l + 1;
      end
      else begin
         while not eoln(afile) do begin
            read(afile, ch);
            c := c + 1;
            s := s + 1;
            if s > buffersize then begin
               s := 1;
            end;
            buffer[s] := ch;
         end;
      end;
   end;

   rewrite(bfile); 
   s := 0;
{
   bufferstart := 0;
}
   while s < bufferstart do begin
      s := s + 1;
      write(output,buffer[s]);
      write(bfile,buffer[s]);
   end;
   writeln(output);
   writeln(bfile)

end;
(* end module bigbuf.themain *)

begin
   themain(afile, bigbufp, bfile);
1: end.
