program dem(demp, list, output);
(* dem: decimal-match

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.02; (* of dem.p 2013 Sep 09
2013 Sep 09, 1.02: wikipedia link
2013 Sep 09, 1.01: tighten code
2013 Sep 09, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.dem *)
(*
name
   dem: decimal-match

synopsis
   dem(demp: in, list: out, output: out)

files

   list:  output of the program

   demp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

description

   Given two decimal numbers report the number of decimal places to
   which they match.

examples

theoretical    =  0.69314718 (0.6931) theoretical maximum efficiency
measured       =  0.69300792 (0.6930) measured efficiency
delta          = -0.00013926 delta of efficiency
                    ***
places identical:     3
identical_digits__1   3.9            (to  1 decimal)
identical_digits_10   3.8561736092   (to 10 decimals)

documentation

see also

   numcode.p

   http://en.wikipedia.org/wiki/Significant_figures

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.dem *)

(* more constants *)
{
   errortolerance = 1e-8; (* for testing *)
}
   errortolerance = 1e-15;

var
   demp, (* file used by this program *)
   list: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module countdigits *)
function countdigits(a,b: real; errortolerance: real): integer;
(* Count the identical digits of numbers a and b.
These are both between 0 and 1.
Counting goes until the error tolerance is reached, where the
error tolerance is also a number between 0 and 1. *)
var
  achop: real; (* the number a chopped up *)
  bchop: real; (* the number b chopped up *)
  atrunc: integer; (* trunc of achop *)
  btrunc: integer; (* trunc of bchop *)
  et: integer; (* number of digits in error tolerance *)
  equal: boolean; (* the digits are still equal *)
  pindex: integer; (* digit count *)
begin
  et := round(-ln(errortolerance)/ln(10));
  pindex := 0;
  achop := abs(a);
  bchop := abs(b);
  equal := true;
  while equal and (pindex < et) do begin
     pindex := succ(pindex);
     achop := 10*achop;
     bchop := 10*bchop;
     atrunc := trunc(achop);
     btrunc := trunc(bchop);
     {writeln(output, atrunc:1,' ', btrunc:1);}
     if atrunc <> btrunc then begin
        equal := false;
     end;
  end;
  countdigits := pindex - 1;
end;
(* end module countdigits *)

(* begin module dem.themain *)
procedure themain(var demp, data: text);
(* the main procedure of the program *)
var
   parameterversion: real; (* parameter version number *)
   theoretical: real; (* theoretical value *)
   measured: real;    (* measured value *)
   delta: real;       (* difference between measured and theoretical values *)
   places: integer; (* how close measured efficiency matches maximum
      in decimal places *)
   pindex: integer; (* index to digits of places *)
   identicaldigits: real; (* computed number of places as a real number *)

begin
   writeln(output,'dem ',version:4:2);
   reset(demp);
   readln(demp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   readln(demp, theoretical);
   readln(demp, measured);

   rewrite(data);
   writeln(data,'dem ',version:4:2);
   writeln(data,'theoretical    =  ', theoretical:10:8,
                                      ' (',theoretical:6:4,')',
                                      ' theoretical maximum efficiency');

   writeln(data,'measured       =  ', measured:10:8,
                                      ' (', measured: 6:4,')',
                                      ' measured efficiency');

   delta := measured - theoretical;
   write  (data,'delta          = ');
   if delta >= 0 then write(data,'+');
   writeln(data, delta:10:8, ' delta of efficiency');
   write  (data,'                  ','  ');
   (* places := trunc(-ln(abs(delta))/ln(10)); This doesn't work well *)
   places := countdigits(measured, theoretical, errortolerance);

   for pindex := 1 to places do write(data,'*');
   writeln(data);
   writeln(data, 'places identical:  ',' ':(places),places:1);
   if abs(delta) > 0
   then identicaldigits := abs(ln(abs(delta))/ln(10))
   else identicaldigits := 0;
   writeln(data, 'identical_digits__1',' ':(places), identicaldigits:3:1,
           ' ':9, '   (to  1 decimal)');
   writeln(data, 'identical_digits_10',' ':(places), identicaldigits:10:10,
                     '   (to 10 decimals)');
   if places = 0
   then begin
      writeln(data, 'no places match exactly')
   end;

end;
(* end module dem.themain *)

begin
   themain(demp, list);
1: end.
