program evenco(input, output);
(* evenco: even column - read integers from a file and output with leading spaces

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.12; (* of evenco.p 2013 Apr 13
2013 Apr 13, 1.12 brush up documentation
2013 Apr 12, 1.11 origin from 1.10 of number.p 2010 Mar 01
*)
(* end module version *)

(* begin module describe.evenco *)
(*
name
   evenco: even column - read integers from a file and output with leading spaces

synopsis
   evenco(input: in, output: out)

files
   input:  integers, one per line
   output: input file with leading spaces

description

   Add leading spaces to integers.  This makes them right justified.

examples

documentation

see also
   number.p
   cumulative.p

author
   Thomas Dana Schneider

bugs

technical notes

   Note from number.p:
   Perhaps the program should have an option to define the number of columns
   for the line numbers.  An earlier version computed this.  However, to do
   that requires resetting the file and prevents the program from being used
   as part of a pipe.  The current program simply uses a default value for
   the number of columns so that number can be used as part of a pipe.

   See cumulative.p for a solution, not yet implemented.

*)
(* end module describe.evenco *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module evenco.themain *)
procedure themain(var fin,fout: text);
(* the main procedure of the program *)
{
const
   decimals = 5; (* that should be sufficient for most files! *)
   decimals = 1; (* No blanks before numbers! *)
}
var
   count: integer; (* count of the lines produced so far *)
   decimals: integer; (* decimal places to express the value of lines in *)
   internal: text; (* internal file to store the integers to allow
     this to be used in a pipe! *)
   lines: integer; (* total lines in the file *)
   i: integer; (* an integer read in *)
   maxi: integer; (* largest value of i *)
begin
   count := 0;
   maxi := 0;
   rewrite(internal);
   while not eof(fin) do begin
      count := succ(count);
      readln(fin, i);
      writeln(internal, i:1);
      if maxi < i then maxi := i;
   end;
   lines := count;

   rewrite(fout);
   if lines > 0 then begin
      (* determine how many decimal places to put out for line numbers *)
      decimals := trunc(ln(maxi)/ln(10))+1;
      reset(internal);
      for count := 1 to lines do begin
         readln(internal,i);
         writeln(fout,i:decimals);
      end;
   end

end;
(* end module evenco.themain *)

begin
   themain(input,output);
1: end.

{suppressed code from number.p:
 this is crazy.  the extra reset prevents
this from being used in pipes.  
   (* determine number of lines in the file *)
   lines := -1; (* don't count the line that has the eof on it! *)
   reset(fin);
   while not eof(fin) do begin
      lines := lines + 1;
      readln(fin)
   end;
   if lines > 0 then begin
      (* determine how many decimal places to put out for line numbers *)
      decimals := trunc(ln(lines)/ln(10))+1;
      (* copy the file *)
      reset(fin);
      for count := 1 to lines do begin
         write(fout,count:decimals,' ');
         copyaline(fin,fout)
      end;
   end;
}
