program fl(input,output);
(* fl: a program to find and list from the input file lines with
a length of greater than maxline characters
      john hoffhines and tom schneider
      module libraries: delman, delmods *)

label 1; (* end of flag *)

(* begin module version *)
const version = 1.14; (* of fl.p 1995 Sep 21
origin: of fl 1995 september 21
origin: of flag 1982 august 15 *)
(* end module version *)

(* begin module describe.flag *)
(*
name
      fl: points out excessively long lines

synopsis
      flag (input: in, output: out)

files
      input: a text file; typically pascal source code.
      output: the first line of input followed by a list of the lines which are
            too long.  the list gives the line number of each line, the line
            itself and a flag on the last acceptable character.
description
      during transportation of files from one computer to another, lines
      longer than 80 characters are often truncated to 80 characters to make
      'card images' on the tape.  this byzantine practice is left over
      from the days when cards were the state-of-the-art in talking to
      computers.  since the tape does not know what a 'card image' is,
      and since cards are going the way of the passenger pigeon,
      this is like equipping a nuclear oil tanker with oars.  maybe
      someday things will be different, but until then, flag exists
      and will detect long lines, allowing one to fix a program or file
      before transportation.  note: trailing blanks on each line are
      ignored.

      This program is like the flag program except that it works from
      the standard input and output files.

author
      john hoffhines and tom schneider

see also
      flag.p

bugs
      none known

technical notes
      the constant 'maxline' defines the number of characters accepted
      on each line.  we recommend that maxline be set to 80 because this
      is the standard number of characters on a punched card.

*)
(* end module describe.flag *)

(* more constants *)
      maxline = 80;  (* lines longer than maxline are sent to output *)
      maxstring = 3000; (* longest line flag will read *)

type  string = record (* for storing and checking characters in strings *)
               letters : array[1..maxstring] of char;
               length: integer;
               end;

var
     countlines, (* the number of the line being read *)
     errors: integer; (* number of lines longer than maxline *)
     buffer: string; (* storage form of input data *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.51 85 apr 17 tds/gds' *)

procedure checkin(var thefile: text);
(* checks if thefile is empty *)
      begin (* checkin *)
         if eof(thefile) then begin
            writeln(output,'i am frightfully sorry old bean but if you',
                           ' want me to');
            writeln(output,'function properly you must not feed me',
                           ' empty files.');
            writeln(output);
            halt
         end
      end; (* checkin *)

procedure readline(var thefile: text; var aline: string);
(* read aline from thefile *)

      begin (* readline *)
         aline.length := 0;
         while not eoln(thefile) do begin
            aline.length := succ(aline.length);
            if aline.length > maxstring then begin
               writeln(output,'a line exceeds ',maxstring:1,
                              ' characters, increase constant maxstring');
               halt
            end;
            read(thefile,aline.letters[aline.length]);
         end;
         readln(thefile)
      end; (* readline *)

procedure untail (var astring: string);
(* removes blanks on the tail of a string *)

      begin (* untail *)
         if astring.length > 1
         then while (astring.letters[astring.length] = ' ') and
            (astring.length > 1)
         do astring.length := pred(astring.length)
      end; (* untail *)

procedure writeline(var thefile: text;
                    var aline: string);
(* write aline to the file *)

var  count: integer;  (* the number of characters in aline *)

      begin (* writeline *)
         count := 1;
         while count <= aline.length do begin
            write(thefile,aline.letters[count]);
            count := succ(count)
         end;
         writeln(thefile)
      end; (* writeline *)

begin (* flag *)
      writeln(output,' flag ',version:4:2);
      writeln(output,'  lines that exceed ',maxline:1, ' characters');
      writeln(output);

      checkin(input);

      readline(input,buffer);
      writeline(output,buffer);
      (* the above mini-procedure places the first line of the input file
      at the beginning of output.  this line is typically the
      name of the program *)

      reset(input);
      countlines := 0; errors := 0;

      while not eof(input) do begin
         readline(input, buffer);
         countlines := countlines + 1;
         untail(buffer);
         if buffer.length > maxline then begin
            errors := errors + 1;
            writeln(output);
            writeln(output,'line number ',countlines:1);
            writeline(output,buffer);
            writeln(output,' ':(maxline-1),'^')
         end;
      end;

      writeln(output,' ',errors:1,' lines are over ',
                     maxline:1,' characters long');
      if errors > 0 then begin
         writeln(output);
         writeln(output,' this file may be difficult to transport');
         writeln(output)
      end;
1: end. (* flag *)
