program httplink({httplinkp,} input, output);
(* httplink: find URLs in a file and make them into links

  Tom Schneider
  permanent email: toms@alum.mit.edu
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms/


 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.05; (* of httplink.p 1999 August 2
origin 1999 July 29 *)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.httplink *)
(*
name
   httplink: find URLs in a file and make them into links

synopsis
   httplink(afile: in, httplinkp: in, input: out, output: out)

files
   input:  a file.

   httplinkp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.
(not used)

   output: a copy of the input file with URLs turned into html links.

description
   Convert a file to html links.

examples

documentation

see also

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.httplink *)

(* LOCK begin module string.const *)
      maxstring = 1000; (* the maximum string *)
(* LOCK end module string.const version = 4.36; (@ of prgmod.p 1999 July 22 *)

(* begin module filler.const *)
      fillermax = 50; (* the size of the filler array for a string *)
(* end module filler.const version = 4.36; (@ of prgmod.p 1999 July 22 *)
 
type

(* begin module interact.type *)
(* begin module string.type *)
   stringptr = ^string; (* pointer to a string *)
   string = record (* a string of characters *)
      letters: array[1..maxstring] of char; (* the letters in the string *)
      length: integer; (* the number of characters in the string *)
      current: integer; (* the letter we are working on *)
      next: stringptr; (* the next string in a series *)
   end;
(* end module string.type version = 4.36; (@ of prgmod.p 1999 July 22 *)
(* end module interact.type version = 4.36; (@ of prgmod.p 1999 July 22 *)

(* begin module trigger.type *)
      trigger = record (* an object to be searched for *)
         seek: string; (* the characters looked for *)
         state: integer; (* how close to triggering we are *)
         skip: boolean; (* trigger not found- skip the line *)
         found: boolean (* the trigger was found *)
      end;
(* end module trigger.type version = 4.36; (@ of prgmod.p 1999 July 22 *)

(* begin module filler.type *)
      (* the following is an array used to fill a string.
      it is convenient to have it much shorter than the maxstring, so that
      it is easy to fill the string using procedure fillstring.
      the user must declare the value of constant fillermax. *)
      filler = packed array[1..fillermax] of char;
(* end module filler.type version = 4.36; (@ of prgmod.p 1999 July 22 *)

{
var
   httplinkp: text; (* file used by this program *)
}

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 4.36; (@ of prgmod.p 1999 July 22 *)

(* begin module clearstring *)
procedure clearstring(var ribbon: string);
(* empty the string *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
      with ribbon do begin
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0;
         current := 0;
      end
end; (* clearstring *)
procedure initializestring(var ribbon: string);
(* start the string with a nil pointer.   This routine
should be called before doing linked list work.  This allows
the standard string routines to clear the string without
killing the pointer. *)
begin (* initializestring *)
   clearstring(ribbon);
   ribbon.next := nil;
end; (* initializestring *)
(* end module clearstring version = 4.36; (@ of prgmod.p 1999 July 22 *)

(* begin module filler.fillstring *)
procedure fillstring(var s: string; a: filler);
(* this procedure makes it reasonably easy to fill the string s with
characters.  one calls the procedure as: *)
(*                           1         2         3         4         5 *)
(*                  12345678901234567890123456789012345678901234567890 *)
(*   fillstring(s, 'this-is-the-string                                ');
the two comments make it easy to line the characters up. also, for this
example, it was assumed that the length of filler as defined by the
constant fillermax was 50. *)
var
      length: integer; (* of the string without trailing blanks *)
      index: integer; (* of s *)
begin (* fillstring *)
      clearstring(s);
      length := fillermax;
      while (length > 1) and (a[length] = ' ') do length := pred(length);
      if (length = 1) and (a[length] = ' ') then begin
         writeln(output, 'fillstring: the string is empty');
         halt
      end;

      for index := 1 to length do s.letters[index] := a[index];
      s.length := length;
      s.current := 1
end; (* fillstring *)
(* end module filler.fillstring version = 4.36; (@ of prgmod.p 1999 July 22 *)

(* begin module filler.filltrigger *)
procedure filltrigger(var t: trigger;
                      a: filler);
(* fill the trigger t *)
begin (* filltrigger *)
         fillstring(t.seek,a)
end; (* fillstring *)
(* end module filler.filltrigger version = 4.36; (@ of prgmod.p 1999 July 22 *)
(* begin module trigger.proc *)
(* this module allows one to scan a series of characters, as from
an array or a file, and to "trigger" or detect a simple string
in the series.  the advantage of the trigger is that several triggers
can "observe" a stream of characters at once, each looking for a
different thing.
some other modules required: interact.const, interact.type *)

procedure resettrigger(var t: trigger);
(* reset the trigger to ground state *)
begin (* resettrigger *)
      with t do begin
         state := 0;
         skip := false;
         found := false
      end
end; (* resettrigger *)

procedure testfortrigger(ch: char; var t: trigger);
(* look at the character ch.
   if it is part of the trigger (at the current trigger state),
       then the trigger state goes higher.
   if it is not part of the trigger then the trigger state is reset,
      skip is true and one should skip onward to find the trigger.
   if the trigger is found, found is true.
1996 Sep 12: Bug found!  In the case of a trigger "ab", the program
used to miss it for situations like "aab".  This was because at the
first a it would step up.  Then it would see the second a and recognize
that was not part of ab.  It would fail to realize that it could be
the start of a new one.  The code now accounts for that possibility. *)
begin (* testfortrigger *)
      with t do begin
         state := succ(state);
{
writestring(list,seek);
writeln(list,'testfortrigger seek.letters[',state:1,']:',
seek.letters[state],' ch:',ch);
}
         if seek.letters[state] = ch
         then begin
            skip := false;
            if state = seek.length then found := true
                                   else found := false
         end
         else begin
            (* it failed.  But wait!  It could be the beginning
            of a NEW trigger string! *)
            if seek.letters[1] = ch then begin
               state := 1;
               skip := false;
               found := false
            end
            else begin (* reset trigger *)
               state := 0;
               skip := true;
               found := false
            end
         end
      end
end; (* testfortrigger *)
(* end module trigger.proc version = 4.36; (@ of prgmod.p 1999 July 22 *)

(* begin module interact.getstring *)
procedure getstring(var afile: text; var buffer: string;
                    var gotten: boolean);
(* get a line (as a string) from a file not using string calls.  this lets
one obtain lines from a file without interactive prompts *)
var   index: integer; (* of buffer *)
begin (* getstring *)
      clearstring(buffer);
      if eof(afile)
      then gotten := false
      else begin
         index := 0;
         while (not eoln(afile)) and (index < maxstring) do begin
            index := succ(index);
            read(afile, buffer.letters[index])
         end;
         if not eoln(afile) then begin
            writeln(output, ' getstring: a line exceeds maximum string size (',
                            maxstring:1,')');
            halt
         end;
         buffer.length := index;
         buffer.current := 1;
         readln(afile);
         gotten := true
      end
end; (* getstring *)
(* end module interact.getstring version = 4.36; (@ of prgmod.p 1999 July 22 *)

(* begin module httplink.themain *)
procedure themain{(var httplinkp: text)};
(* the main procedure of the program *)
var
   aline: string; (* a line being transferred from input to output *)
   c: char; (* character being transferred from input to output *)
   debugging: boolean; (* show what we are doing*)
   done: boolean; (* done looking for end of URL? *)
   gotten: boolean; (* got aline? *)
   http: trigger; (* trigger for making link *)
   iw: integer; (* writeing index to aline *)
   ir: integer; (* reading index to aline *)
   j: integer; (* index to URL *)
   start: integer; (* start of a URL in aline *)
   stop: integer; (* stop of a URL in aline *)
{
   parameterversion: real; (* parameter version number *)
}
begin
   writeln(output,'<!-- httplink ',version:4:2,' -->');
{
   reset(httplinkp);
   readln(httplinkp, parameterversion);
   if parameterversion < updateversion then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;
}

(*                         1         2         3         4         5 *)
(*                12345678901234567890123456789012345678901234567890 *)
filltrigger(http,'http://                                           ');

debugging := true;
debugging := false;

   while not eof (input) do begin
      getstring(input, aline, gotten);
      resettrigger(http);
      if gotten then {with aline do} begin
         ir := 1;
         iw := 1;
         while ir <= aline.length do begin
            c := aline.letters[ir];
            testfortrigger(c, http);
            if http.found then begin
               start := ir-http.seek.length+1;
if debugging then begin
writeln(output,'!');
writeln(output,'aline.length = ',aline.length:1);
writeln(output,'http.seek.length = ',http.seek.length:1);
writeln(output,'start = ',start:1);
end;
               (* find end *)
               done := false;
               while not done do begin
                  if (aline.letters[ir] = ' ') or
                     (aline.letters[ir] = ',') or
                     (aline.letters[ir] = ')') or
                     (aline.letters[ir] = ']') or
                     (aline.letters[ir] = ':')
                  then begin
                     done := true;
                     stop := ir-1;
                  end
                  else if ir = aline.length then begin
                     stop := ir;
                     ir := succ(ir); (* force conclusion *)
                     done := true;
                  end
                  else ir := succ(ir)
               end;
if debugging then begin
writeln(output,'stop = ',stop:1);
write(output,'captured: "');
for j := start to stop do write(output,aline.letters[j]);
writeln(output,'"');
end;
               (* catch up the output, which was at ir *)
{
write(output,'catchup:"');
}
               while iw < start do begin
                  write(output,aline.letters[iw]);
                  iw := succ(iw)
               end;
{
writeln(output,'"');
}

               (* write out the url as a link *)
               write(output,'<a href = "');
               for j := start to stop do write(output,aline.letters[j]);
               write(output,'">');
               for j := start to stop do write(output,aline.letters[j]);
               write(output,'</a>');
               iw := stop + 1; (* catch up output *)
            end
            else begin
{
               write(output,c);
}
               ir := succ(ir);
            end
         end;
         (* finish line if needed *)
         while iw <= aline.length do begin
            write(output,aline.letters[iw]);
            iw := succ(iw)
         end;
         writeln(output);
      end;
   end

end;
(* end module httplink.themain *)

begin
   themain{(httplinkp)};
1: end.
