program makeinst(instA, instB, makeinstp, dcout, output);
(* makeinst: generate delila instructions for set A and B from dcout.

  Ryan Shultzaberger

in the lab of:

  Dr. Thomas D. Schneider
  toms@alum.mit.edu
  permanent email: toms@alum.mit.edu (use only if first address fails)
  https://alum.mit.edu/www/toms/

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.01; (* of makeinst.p 2005 Oct 13
2005 Oct 13, 1.01: touch up documentation (TDS)
2005 Oct 07, 1.00: upgrade address
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.makeinst *)
(*
name
   makeinst: generate delila instructions for set A and B from dcout.

synopsis
   makeinst(instA: in, instB: in, makeinstp: in, dcout: out, output: out)

files

   instA:  Delila instructions for type A sites.

   instB:  Delila instructions for type B sites. 

   dcout:  Output of the diri program that reports coordinates of A and B and
           information about their individual strengths and their combined
           strength.

   makeinstp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      From To Range (2 integers): This is the From To range to use in the
         inst.  It should be given like "-10 +10".  This will correspond to an
         inst that is "get from 100 -10 to same +10;".

   output: messages to the user

description

   This program will generate two inst files "instA" and "instB" from
   the dcout file created by the diri program.  The makeinst program
   assumes that A is the more upstream of the two binding components
   and assigns the orienation of the instruction on the relative
   positioning of B to A.  That is if A > B then the instruction for
   that pair will have a positive orientation.  If B > A then the
   instruction for that pair will have a negative orientation.  Since
   the header information is not present in dcout it is not reported
   in the inst files.  This must be added later.

examples

   example makeinstp file:

1.00  version of makeinst that this parameter file is designed for.
-100 +100  From/To range to use in inst.


documentation

see also

  diri.p

  delila.p

  {example parameter file: } makeinstp

author

   Ryan Shultzaberger

bugs

technical notes

*)
(* end module describe.makeinst *)

var
   instA, (* file used by this program *)
   instB, (* file used by this program *)
   makeinstp, (* file used by this program *)
   dcout: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module makeinst.themain *)
procedure themain(var instA, instB, makeinstp, dcout: text);
(* the main procedure of the program *)
var
   parameterversion: real; (* parameter version number *)
   fromrange: integer; (* the from range for the inst files *)
   torange: integer; (* the two range from the inst files *)
   coordA, coordB: integer; (* the coordinates of sites A and B *)
   diff: integer; (* the difference between coordA and coordB *)
   c: char; (* first character on the line *)
begin
   writeln(output,'makeinst ',version:4:2);
   reset(makeinstp);
   readln(makeinstp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;
   readln(makeinstp, fromrange, torange);

   reset(dcout);
   rewrite(instA);
   rewrite(instB);
   while not eof(dcout) do begin
      read(dcout,c);
      if c <> '*' then begin
         readln(dcout,coordA, coordB);
         diff:= coordA - coordB;
         if diff >= 0 then begin
            writeln(instA,'get from ',coordA:1,' ',fromrange:1,
                          ' to same +',torange:1,';');
            writeln(instB,'get from ',coordB:1,' ',fromrange:1,
                          ' to same +',torange:1,';');
         end;
         if diff < 0 then begin
            writeln(instA,'get from ',coordA:1,' +',-fromrange:1,
                          ' to same -',torange:1,';');
            writeln(instB,'get from ',coordB:1,' +',-fromrange:1,
                          ' to same -',torange:1,';');
         end;
      end
      else readln(dcout);
   end;

end;
(* end module makeinst.themain *)

begin
   themain(instA, instB, makeinstp, dcout);
1: end.
