program meri(ribl5, ribl3, merip, ribl53, output);
(* meri: merge ribl files

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.15; (* of meri.p 2015 Jul 29
2015 Jul 29, 1.15: work on numbering equations ... not done
2015 Jul 29, 1.14: clear and read ribl5 and ribl3 separately so they
                   can be the same file
2015 Jul 29, 1.13: crash on two copies of lambdacicro.ribl: clear spare1&2!!
2011 Feb 08, 1.12: upgrade documentation, add example ribl
2011 Feb 08, 1.11: add more components to the ribl; stdev?
2011 Feb 08, 1.10: add more components to the ribl; mean = Rseq; initial stdev
2011 Feb 08, 1.09: add more components to the ribl; averageRi
2011 Feb 08, 1.08: add more components to the ribl; bounds
2011 Feb 08, 1.07: add more components to the ribl; anticonsensus
2011 Feb 08, 1.06: add more components to the ribl; consensus
2011 Feb 08, 1.05: display coordinates; create header name; + for fusion
2011 Feb 08, 1.04: clean up; display coordinates
2011 Feb 08, 1.03: read parameters; fill new matrix with 5' matrix; appears right!
2011 Feb 07, 1.02: read parameters; fill new matrix with 5' matrix
2011 Feb 07, 1.01: construct code, now reading and writing ribls
2011 Feb 07, 1.00: origin, has modules
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.meri *)
(*
name
   meri: merge ribl files

synopsis
   meri(ribl5, ribl3: in, merip: in, ribl53: out, output: out)

files

   ribl5, ribl3:  two ribl matricies to obtain parts of

   ribl53:  parts of the two ribl matricies fused together

   merip:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      ribl53name (string): name of the new matrix.  However, if the
      first string on the line is '+', then the name is made by fusing
      the  names of the ribl5 and ribl3.

      ribl5from and ribl5to (integer): from and to locations of the 5'
         ribl to use.

      ribl3from and ribl3to (integer): from and to locations of the 3'
         ribl to use.

      zero (integer): new zero coordinate on the 5' ribl coordinate system.
         This is counted from the 5' end (as 1)

   output: messages to the user

description

   Merge two ribl weight matricies together.  The upstream one is
   called the 5' and the downstream one is called the 3'.

examples

   Fuse the donor splice site (-3 to +6) on the 5' end to the acceptor
   splice site (-25 to +2) on the 3' end by taking the intron side. 
   By our original definition, the zero base is the one on the intron
   side, so the zero coordinates are removed to give numbering:

   -3 -2 -1       from donor
            +1 +2 from acceptor
   -2 -1  0 +1 +2 new numbering using position 3 as the new zero

   The merip parameter file for this is:

1.00  version of meri that this parameter file is designed for.
doac  ribl53name: string name of the merged ribl
-3 -1 ribl5from, ribl5to: range to use from the 5' ribl
+1 +2 ribl3from, ribl3to: range to use from the 3' ribl
3     zero (integer): new zero coordinate on the 5' ribl coordinate system.

documentation

@article{Stephens.Schneider1992,
author = "R. M. Stephens
  and T. D. Schneider",
title = "Features of spliceosome evolution and function
inferred from an analysis of the information at human splice sites",
journal = "J. Mol. Biol.",
volume = "228",
pages = "1124-1136",
note = "\htmladdnormallink
{https://alum.mit.edu/www/toms/papers/splice/}
{https://alum.mit.edu/www/toms/papers/splice/}",
year = "1992"}

see also

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.meri *)

(* begin module string.const *)
      maxstring = 2000; (* the maximum string *)
(* end module string.const version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module info.const *)
   infofield = 10; (* size of field for printing information in bits *)
   infodecim = 6; (* number of decimal places for printing information *)
       (* these are used for conlist only *)
   nfield = 8; (* size of field for printing n, the number of sites *)
(* end module info.const version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module ribl.const module *)
   minribl = -2000; (* lowest  ribl matrix from allowed *)
   maxribl = +2000; (* highest ribl matrix to   allowed *)
   defnegativeinfinity = -1000; (* default for negative infinity
                                   for the Ri(b,l) table *)
(* end module ribl.const version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

type

(* begin module wave.type *)
      waveptr = ^waveparam; (* to link several wave definitions together *)
      waveparam = record (* parameters to define a cosine wave *)
         (* define a cosine wave: *)
         extreme: char; (* h or l, the high or low extreme to be defined *)
         wavelocation: real; (* the location in bases of the extreme *)
         wavebit: real; (* the location in bits of the extreme *)
         waveamplitude: real; (* the amplitude of the wave in bits *)
         wavelength: real; (* the wave length of the wave in bases *)
{ 2000 Nov 1.  This is difficult because the length of a curve is not
simply multiplied by a scale factor.
         BASEdashon: real; (* dashon interval in bases *)
         BASEdashoff: real; (* dashon interval in bases *)
         BASEdashoffset: real; (* dashon interval in bases *)
}
         dashon: real; (* the size of on dashes in cm.
                          dashon <= 0 means no dashes *)
         dashoff: real; (* the size of on dashes in cm *)
         dashoffset: real; (* the size of off dashes in cm *)
         thickness: real; (* thickness of the wave in cm.  <= 0 means default *)
         next: waveptr; (* the next wave *)
      end;
(* end module wave.type version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module base.type *)
   (* define the four nucleotide bases *)
   base = (a,c,g,t);
(* end module base.type version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module string.type *)
   stringptr = ^string; (* pointer to a string *)
   string = record (* a string of characters *)
      letters: array[1..maxstring] of char; (* the letters in the string *)
      length: integer; (* the number of characters in the string *)
      current: integer; (* the letter we are working on *)
      previous: stringptr; (* the previous string in a series *)
      next: stringptr; (* the next string in a series *)
   end;
(* end module string.type version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module scan.type *)
   (* define the ribl matrix.  Source library is prgmod.p
      REQUIRES ribl.const module
      REQUIRES base.type or book.type module
      REQUIRES wave.type module for wave definition! *)
   rblarray = array[a..t, minribl..maxribl] of real; (* real(B,L) *)
   ribltype = record
      riblname: string; (* name of the weight matrix *)
      riblheader: stringptr; (* misc header information, starting with '*' *)
      data: rblarray; (* real(B,L) *)
      numbers: array[a..t, minribl..maxribl] of integer; (* n(b,l) *)
      frombase, tobase: integer; (* range of the matrix *)
      mean, stdev: real; (* mean and standard deviation of distribution *)
      consensus: real; (* Ri value of consensus sequence *)
      anticonsensus: real; (* Ri value of anticonsensus sequence *)
      averageRi: real; (* average Ri for random sequence *)
      n: integer; (* number of sequences used to create matrix *)
      symmetry: char; (* the symmetry of the Riblmatrix *)
      waves: waveptr; (* cosine wave definitions for the matrix *)
      cmperbase, cmperbit: real; (* required for wave dash definition *)
      Ribound: real; (* the Ri boundary for this definition *) 
      Zbound: real; (* the Z boundary for this definition *) 
      Pbound: real; (* the probability boundary for this definition *) 
   end;
   ribltypeptr = ^ribltype; (* pointer to ribltype so it does
      not need to be created unless actually used *)
(* end module scan.type version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

var
   ribl5, ribl3, (* file used by this program *)
   merip, (* file used by this program *)
   ribl53: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module skipblanks *)

(* 2003 July 31: tab is considered a blank character *)

function isblank(c: char): boolean;
(* is the character c blank or tab? *)
const
   tab = 9; (* tab character *)
begin
   isblank := (c = ' ') or (ord(c) = tab)
end;

procedure skipblanks(var thefile: text);
(* skip over blanks until a non-blank, or end of line, is found *)
begin
   while isblank(thefile^) and not eoln(thefile) do get(thefile);
end;

procedure skipnonblanks(var thefile: text);
(* skip over nonblanks until a blank, or end of line, is found *)
begin
   while (not isblank(thefile^)) and not eoln(thefile) do get(thefile);
end;

procedure skipcolumn(var thefile: text);
(* skip over a data column *)
begin
   skipblanks(thefile); skipnonblanks(thefile)
end;
(* end module skipblanks version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module clearstring *)
(* These modules clear strings in various ways *)
(* ---- *)
procedure emptystring(var ribbon: string);
(* empty the contents of the string but do NOT
remove the pointer.  This is useful for clearing
one string within a linked list of them. *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
   with ribbon do begin
      for index := 1 to maxstring do letters[index] := ' ';
      length := 0;
      current := 0;
   end
end; (* emptystring *)
(* ---- *)
procedure clearstring(var ribbon: string);
(* empty the string and remove the pointer.
As of 2010 Dec 04 there is a previous pointer that
allows proper removal of the linked list using disposestring.
This clearstring routine is for the initial clearing
of the variables. *)
begin (* clearstring *)
   with ribbon do begin
      emptystring(ribbon);
      previous := nil;
      next := nil;
   end
end; (* clearstring *)
(* ---- *)
procedure disposestring(var ribbon: stringptr);
(* remove the linked list from the string in revere order *)
var
   t: stringptr; (* a link in the string *)
begin
   t := ribbon;
   (* find the end *)
   while t^.next <> nil do t := t^.next;
   (* move backwards and dispose links *)
   while t^.previous <> nil do begin
      t := t^.previous;
      dispose(t^.next);
   end;
   dispose(t);
end; (* disposestring *)
(* ---- *)
procedure initializestring(var ribbon: string);
(* start the string with a nil pointer.   This routine
should be called before doing linked list work.  This allows
the standard string routines to clear the string without
killing the pointer.  This is now deprecated, do not use it
since clearstring still clears the next pointer. *)
begin (* initializestring *)
   writeln(output,'remove initializestring routine!');
   writeln(output,'replace it with clearstring routine!');
   halt; (* to force deprecation *)
   clearstring(ribbon);
   ribbon.next := nil;
end; (* initializestring *)
(* end module clearstring version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module copystring *)
procedure copystring(a: string; var b: string);
(* copy string a to b.  The length, characters and
where the string points to are all copied. *)
var l: integer; (* index to the string *)
begin
   b.length := a.length;
   for l := 1 to a.length do b.letters[l] := a.letters[l];
   b.current := a.current;
   b.next := a.next;
end;
(* end module copystring version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module concatstring *)
procedure concatstring(var a, b: string; var c: string);
(* concatenate a and b and put in c *)
var   abindex: integer; (* index to both strings *)
      cindex: integer; (* index to both strings *)
begin (* concatstring *)
   emptystring(c);
   c.length := a.length + b.length;
   if c.length > maxstring then begin
      writeln(output,'string too long in concatstring');
      halt
   end;

   cindex := 0;
   for abindex := 1 to a.length do begin
      cindex := succ(cindex);
      c.letters[abindex] := a.letters[abindex]
   end;

   for abindex := a.length+1 to a.length+b.length do begin
      cindex := succ(cindex);
      c.letters[abindex] := b.letters[abindex-a.length]
   end;
end; (* concatstring *)
(* end module concatstring version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module writestring version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module attachcharacterstring *)
procedure attachcharacterstring(c: char; var a: string);
(* attach the character c to the end of a *)
begin
   if a.length > maxstring then begin
      writeln(output,'attachcharacterstring: maxstring exceeded');
      halt
   end;
   with a do begin
      length := succ(length);
      letters[length] := c
   end
end;
(* end module attachcharacterstring version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module readstringline *)
procedure readstringline(var afile: text; var buffer: string);
(* get a line (as a string) from a file. *)
var   index: integer; (* of buffer *)
begin (* getstring *)
   clearstring(buffer);
   index := 0;
   while (not eoln(afile)) and (index < maxstring) do begin
      index := succ(index);
      read(afile, buffer.letters[index])
   end;
   if not eoln(afile) then begin
      writeln(output, 'readstringline: a line exceeds maximum string size (',
                      maxstring:1,')');
      halt
   end;
   buffer.length := index;
   buffer.current := 1;
   readln(afile);
end; (* readstringline *)
(* end module readstringline version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module detabstring *)
procedure detabstring(var s: string);
(* Tabs are converted to blanks.  This allows the tab to be a special "solid"
kind of blank, but the string written won't vary with position on the line
depending on the tabstops. *)
const
   tabcharacter = 9; (* the ordinal of the tab character *)
var
   c: char; (* a character write out *)
   i: integer; (* index to the string s *)
begin
   with s do for i := 1 to length do begin
      c := letters[i];
{
for testing:
      if ord(c) = tabcharacter then c := 'T'; (* tab converted to blank *)
}
      if ord(c) = tabcharacter then c := ' '; (* tab converted to blank *)
      letters[i] := c
   end;
end;
(* end module detabstring version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module readquotestring *)
procedure readquotestring(var afile: text; var s: string);
(* read the quoted string from afile, remove the quotes.  Trailing
blanks are converted to tab characters. *)
const
   tabcharacter = 9; (* the ordinal of the tab character *)
var
   c: char; (* a character read *)
   i: integer; (* index to the string s *)
begin
   clearstring(s);
   skipblanks(afile);
   with s do begin
      read(afile,c);
      if c <> '"' then begin
         writeln(output, 'quote string expected but " missing');
         writeln(output, 'character found instead:"',c,'"');
         halt
      end;
      if eoln(afile) then begin
         writeln(output, 'missing end of quote string " string');
         halt
      end;
      repeat
         read(afile,c);
         if c <> '"' then begin
            length := succ(length);
            letters[length] := c;
         end;
      until (c = '"')
            or eoln(afile) or (length = maxstring);
      if length = maxstring then begin
         writeln(output, 'A quote string exceeded ',maxstring:1,
                         ' characters.');
         writeln(output, 'Use a shorter string or increase constant',
                         ' maxstring.');
         halt
      end;
      if c <> '"' then begin
         writeln(output, 'The second " mark of a string is missing.');
         halt
      end;
      (* convert all blanks to tab characters *)
      for i := 1 to length do begin
         if letters[i] = ' '
         then letters[i] := chr(tabcharacter); (* set to tab character *)
      end;
   end;
end;
(* end module readquotestring version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module writequotestring *)
procedure writequotestring(var afile: text; s: string);
(* Write the string s to afile, surrounded by double quote marks.  Tabs are
converted to blanks.  This allows the tab to be a special "solid" kind of
blank, but the string written won't vary with position on the line depending
on the tabstops. *)
begin
   detabstring(s);
   write(afile, '"');
   writestring(afile, s);
   write(afile, '"');
end;
(* end module writequotestring version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module writestringlines *)
procedure writestringlines(var afile: text; l: stringptr);
(* write lines of strings to a file. *)
var
   h: stringptr; (* pointer to l *)
begin (* getstring *)
   if l <> nil then begin
      h := l;
      while h <> nil do begin
         writestring(afile, h^);
         writeln(afile);
         h := h^.next;
      end;
   end
end;
(* end module writestringlines version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module equalstring *)
function equalstring(a, b: string): boolean;
(* Test for equality between two strings at current positions.
NOTE:  A compiler bug results if one directly tests this way:
   if thedefinition^.nametag = aname
The reason is completely not clear!  I showed that the parts of the
strings were identical, but the whole was not by this test.
For this reason it is *critical to test strings with equalstring.
*)
var   index: integer; (* index to both strings *)
      equal: boolean; (* are letters in a and b the same? *)
begin (* equalstring *)
      if a.length = b.length
      then begin
         index := 1;
         repeat
            equal := (a.letters[index] = b.letters[index]);
            index := succ(index)
         until (not equal) or (index > a.length);
         equalstring := equal
      end
      else equalstring := false
end; (* equalstring *)
(* end module equalstring version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module readheader *)
procedure readheader(var afile: text; ch: char; var theheader: stringptr);
(* Read the header from a file.  Only read lines that begin with ch.
2004 Dec 15: Handle end of file (eof) gracefully. *)
var
   h: stringptr; (* pointer to the riblheader *)
begin
   if eof(afile) then begin
      clearstring(theheader^);
   end
   else begin
      if afile^ = ch then begin
         new(theheader);
         h := theheader;
         while afile^='*' do begin
            readstringline(afile, h^);
            if not eof(afile) then begin
               if afile^='*' then begin
                  new(h^.next);
                  h := h^.next;
               end
               else h^.next := nil;
            end
         end
      end
      else begin
         theheader := nil
      end;
   end;
end;
(* end module readheader version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module grabtoken *)
procedure grabtoken(var thefile: text; var thestring: string);
(* skip any blanks and then grab the next token from the file *)
var
   c: char; (* a character in thefile *)
   done: boolean; (* done finding the name *)
begin
   skipblanks(thefile);
   done := false;
   with thestring do begin
      length := 0;
      while not done do begin
         if eoln(thefile)
         then done := true
         else begin
            read(thefile,c);
            if c = ' '
            then done := true
            else begin
               length := succ(length);
               letters[length] := c;
            end
         end
      end
   end
end;
(* end module grabtoken version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module readwaveparameters *)
procedure readawaveparameter(var afile: text; var wp: waveparam);
{
                                 cmperbase, cmperbit: real);
}
(* Read one wave parameter from a file. *)
{
2000 Nov 1:  The following experimental code is not implemented:
See the coscurve program
(https://alum.mit.edu/www/toms/delila/coscurve.html) for details
about the conversion from bases to cm.  The code now requires
cm per base and cmperbit.
const
   lover2pi = 1.2160067233; (* L/2pi for L being the length of
       the cosine function from 0 to 2pi *)
var
   multiplier: real; (* factor to multiply by to convert from bases to cm. *)
}
begin
   readln(afile,wp.extreme);
   readln(afile,wp.wavelocation);
   readln(afile,wp.wavebit);
   readln(afile,wp.waveamplitude);
   readln(afile,wp.wavelength);
   if wp.wavelength <= 0.0 then begin
      writeln(output,'wave parameters: wavelength must be positive');
      halt
   end;
   (* one may skip the dash and thickness parameters
      of the last wave definition.  This allows
      backwards compatability, but should not be
      used in general. *)
   if eof(afile)
   then begin
      wp.dashon := 0;
      wp.dashoff := 0;
      wp.dashoffset := 0;
      wp.thickness := 0
   end
   else begin
      (* implement new method to read dash information
         without breaking previous wave parameter files *)
      if afile^ <> 'd'
      then begin (* default to old method *)
         readln(afile,wp.dashon);
         wp.dashoff := wp.dashon;
         wp.dashoffset := 0;
      end
      else begin
         get(afile); (* skip the 'd' *)
         (* reading the parameters directly in cm would require
            that the user compute or guess the values.
            So read in as bases and compute cm.
            Unfortunately the length of a curve is not the simple
            scale factor.  Make the user guess for now.
         *)
         readln(afile,wp.dashon, wp.dashoff, wp.dashoffset)

{
experimental code to account for actual wave length:
         readln(afile, wp.BASEdashon, wp.BASEdashoff, wp.BASEdashoffset);
         multiplier    := 1;
         multiplier    := lover2pi
 * wp.waveamplitude (* in bits *)
 * cmperbit (* converts to cm *)
 / wp.wavelength (* 10.6 bp/turn *)
* cmperbase
;
multiplier    := lover2pi;
multiplier    := 1.0;
multiplier    := 1.0/lover2pi;
         multiplier := lover2pi
                       * wp.waveamplitude (* in bits *)
                       * cmperbit (* converts to cm *)
                       * wp.wavelength (* 10.6 bp/turn *)
                       * cmperbase;
         multiplier := 1.0
         wp.dashon     := multiplier * wp.BASEdashon;
         wp.dashoff    := multiplier * wp.BASEdashoff;
         wp.dashoffset := multiplier * wp.BASEdashoffset;
with wp do begin
   writeln(output, 'cmperbase  = ',cmperbase:5:2);
   writeln(output, 'cmperbit   = ',cmperbit:5:2);
   writeln(output, 'dashon     = ',dashon:5:2);
   writeln(output, 'dashoff    = ',dashoff:5:2);
   writeln(output, 'dashoffset = ',dashoffset:5:2);
   writeln(output, 'waveamplitude   = ',wp.waveamplitude:5:2);
   writeln(output, 'wavelength      = ',wp.wavelength:5:2);
end;
}

      end;
      if eof(afile)
      then wp.thickness := 0
      else readln(afile,wp.thickness);
   end;
end;

procedure readwaveparameters(var afile: text; var w: waveptr);
{
                                 cmperbase, cmperbit: real);
}
(* read from afile the wave parameters w.  The routine is done either at end
of file, when a period ('.') appears or at a blank line.  This makes it
compatable with different reading mechanisms. *)
var
   done: boolean; (* done reading *)
   morethanone: boolean; (* more than one wave? *)
   p: waveptr; (* one of the definitions of a cosine wave *)
procedure waystoend;
(* if end of file, a blank line or a period, consider the job done *)
var
   clear: boolean; (* not eof, so clear to test for comment *)
begin
   if eof(afile) then done := true;
   clear := false;
   if not done (* skip any number of comments *)
   then while (not clear) and (not done) do begin
      if eof(afile) then begin
         done := true;
         clear := true;
      end
      else if afile^ = '*'
           then readln(afile)
           else clear := true
   end;
   if not done then if eoln(afile) then done := true;
   if not done then if afile^='.' then begin
      readln(afile);
      done := true;
   end;
end;
begin
   done := false;
   morethanone := false;
   w := nil;
   waystoend;
   if not done then begin
      new(w);
      p := w;
      while not done do begin
         waystoend;
         if done
         then p^.next := nil
         else begin
            if afile^='*'
            then readln(afile) (* skip any number of comments *)
            else begin
               if morethanone then begin
                  new(p^.next);
                  p := p^.next
               end;
{
               readawaveparameter(afile, p^, cmperbase, cmperbit);
}
               readawaveparameter(afile, p^);
               (* if we ever come back again there will be more than one: *)
               morethanone := true;
            end
         end
      end
   end
   else w := nil;
end;
(* end module readwaveparameters version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module writewaveparameters *)
procedure writewaveparameters(var afile: text; w: waveptr);
(* write wave parameters to a file *)
const
   wid = 4; (* width of numbers *)
var
   i: waveptr; (* pointer to the waves in w *)
begin
   i := w;
   while i <> nil do with i^ do begin
      writeln(afile,extreme,
                    '  extreme: char; h or l, the high or low extreme to be defined');
      writeln(afile,wavelocation:wid:1,
                    '  wavelocation: real; the location in bases of the extreme');
      writeln(afile,wavebit:wid:1,
                    '  wavebit: real; the location in bits of the extreme');
      writeln(afile,waveamplitude:wid:1,
                    '  waveamplitude: real; the amplitude of the wave in bits');
      writeln(afile,wavelength:wid:1,
                    '  wavelength: real; the wave length of the wave in bases');
{
      writeln(afile,dash:wid:1,
                    '  dash: real; the size of dashes in cm.',
                    '  dash <= 0 means no dashes');
      writeln(afile,    BASEdashon:wid:1,
                   ' ', BASEdashoff:wid:1,
                   ' ', BASEdashoffset:wid:1,
                   ' dashon, dashoff, dashoffset (in bases)');
}
      writeln(afile,    dashon:wid:1,
                   ' ', dashoff:wid:1,
                   ' ', dashoffset:wid:1,
                   ' dashon, dashoff, dashoffset (in bases)');
      writeln(afile,thickness:wid:1,
                    '  thickness: real; the thickness of the cosine wave.',
                    '  <=0 means default');
      i := i^.next
   end;
   writeln(afile,'.');
end;
(* end module writewaveparameters version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module getriblmatrix *)
procedure getriblmatrix(var afile: text; var matrix: ribltypeptr);
(* Get the matrix from a file, with the defining coordinate limits,
followed by the mean and standard deviation and other data.
The entire matrix is read in. *)
const
   clearvalue = 5.0; (* If a large bit position is ever seen (for DNA), we
      KNOW there is a bug in the code! *)
var
   b: base; (* a base in the matrix *)
   l: integer; (* a coordinate in the matrix *)
   thel: integer; (* the l value in the file (should match l) *)

   teststring: string; (* string to test for 'ribl ' at matrix start *) 
   riblstart: string; (* word at the start of the matrix *)

procedure testforend;
(* test for end of file *)
begin
   if eof(afile) then begin
      writeln(output,'part or all of the ribl matrix is missing');
      halt;
   end
end;
procedure skip;
(* skip comment lines *)
begin
   testforend;
   while afile^='*' do begin
     readln(afile);
     testforend
   end
end;
begin
   new(matrix);
   with matrix^ do begin

      (* read name of the matrix *)
      grabtoken(afile, riblstart);
      teststring.letters[1] := 'r'; 
      teststring.letters[2] := 'i';
      teststring.letters[3] := 'b';
      teststring.letters[4] := 'l';
      teststring.length := 4; 
 
      if not equalstring(riblstart, teststring) then begin
         writeln(output,'The ribl matrix must begin with ''ribl "name"''');   
         writequotestring(output,riblstart);
         writeln(output, ' seen instead');
         halt
      end;
 
      skipblanks(afile); (* skip any spaces *)
      readquotestring(afile, matrix^.riblname);
      readln(afile);

      readheader(afile,'*',riblheader);

      readln(afile,frombase,tobase);

      if frombase < minribl then begin
         writeln(output,'Matrix is too large for program.');
         writeln(output,'frombase (=',frombase:1,')');
         writeln(output,'minribl (=',minribl:1,')');
         writeln(output,'Decrease minribl or increase frombase');
         halt
      end;

      if tobase > maxribl then begin
         writeln(output,'Matrix is too large for program.');
         writeln(output,'tobase (=',tobase:1,')');
         writeln(output,'maxribl (=',maxribl:1,')');
         writeln(output,'Increase maxribl or decrease tobase');
         halt
      end;

      (* clear unused matrix for safety *)
      for l := minribl to frombase-1
      do for b := a to t do matrix^.data[b,l] := clearvalue;
      for l := tobase+1 to maxribl
      do for b := a to t do matrix^.data[b,l] := clearvalue;

      for l := frombase to tobase do begin
         testforend;
         for b := a to t do read(afile,matrix^.data[b,l]);
         read(afile,thel);
         if thel <> l then begin
            writeln(output, 'positions in ribl do not match from-to range');
            halt;
         end;
         for b := a to t do read(afile,matrix^.numbers[b,l]);
         readln(afile)
      end;

      skip; readln(afile,mean);
      skip; readln(afile,stdev);
      skip; readln(afile,consensus);
      skip; readln(afile,anticonsensus);
      skip; readln(afile,averageRi);
      skip; readln(afile,n);
      skip; readln(afile,symmetry);
      skip; readln(afile,Ribound);
      skip; readln(afile,Zbound);
      skip; readln(afile,Pbound);
{this is a guess - set the converters to neutral for now
      cmperbase := 1.0;
      cmperbit := 1.0;
      readwaveparameters(afile, waves, cmperbase, cmperbit);
}
      readwaveparameters(afile, waves);
   end;
end;
(* end module getriblmatrix version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module putriblmatrix *)
procedure putriblmatrix(var afile: text;
                    matrix: ribltypeptr);
(* write the ribl matrix to afile. *)
var
   lindex: integer; (* index to Riblmatrix, equivalent to l *)
begin
   if matrix <> nil then begin
      with matrix^ do begin
         write(afile,'ribl');
         write(afile,' ');
         writequotestring(afile, riblname);
         writeln(afile);
         (* write the header *)
         writestringlines(afile, riblheader);
         writeln(afile, frombase: nfield,' ', tobase: nfield,' frombase, tobase');
         for lindex := frombase to tobase do begin
            write(afile,' ',
                       ' ',data[a, lindex]:infofield:infodecim,
                       ' ',data[c, lindex]:infofield:infodecim,
                       ' ',data[g, lindex]:infofield:infodecim,
                       ' ',data[t, lindex]:infofield:infodecim,
                       ' ',lindex:nfield);
            write(afile,' ',numbers[a,lindex]:nfield,
                        ' ',numbers[c,lindex]:nfield,
                        ' ',numbers[g,lindex]:nfield,
                        ' ',numbers[t,lindex]:nfield);
            writeln(afile);
         end;
         writeln(afile,'*');
         writeln(afile,mean:infofield:infodecim, 
                       ' bits = mean (Rsequence of selected region)');
         writeln(afile,stdev:infofield:infodecim,
                       ' bits = standard deviation');
         writeln(afile,'*');
         writeln(afile,consensus:infofield:infodecim,
                       ' bits = Ri of consensus sequence',
                       ' from ',frombase:1,' to ',tobase:1);
         writeln(afile,anticonsensus:infofield:infodecim,
                       ' bits = Ri of anticonsensus sequence',
                       ' from ',frombase:1,' to ',tobase:1);
         writeln(afile,'*');
         writeln(afile,averageRi:infofield:infodecim,
                       ' bits = average Ri for random sequence',
                       ' from ',frombase:1,' to ',tobase:1);
         writeln(afile,'*');
         writeln(afile,n:1,
                       ' n, number of sequences used to create the matrix');
         writeln(afile,'*');
         case symmetry of
           'a': write(afile, 'asymmetric');
           'e': write(afile, 'even');
           'o': write(afile, 'odd');
         end;
         writeln(afile, ' symmetry of the matrix');
         writeln(afile,'*');
         writeln(afile,Ribound:infofield:infodecim,
                       ' Ri bound: lower bound on Ri');
         writeln(afile,Zbound:infofield:infodecim,
                       ' Z bound: lower bound on Z');
         writeln(afile,Pbound:infofield:infodecim,
                       ' P bound: upper probability');
         writeln(afile,'*');
         writewaveparameters(afile, waves);
         writeln(afile);
      end
   end
end;
(* end module putriblmatrix version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module between *)
function between(a,b,c: real):boolean;
(* is b between a and c? *)
begin between:=(a<=b) and (b<=c) end;
(* end module between version = 5.46; (@ of prgmod.p 2011 Jan 20 *)

(* begin module basetochar *)
function basetochar(ba:base):char;
(* convert a base into a character *)
begin
      case ba of
         a: basetochar:='a';
         c: basetochar:='c';
         g: basetochar:='g';
         t: basetochar:='t';
      end
end;
(* end module basetochar from delmod by copy *)

(* begin module meri.themain *)
procedure themain(var ribl5, ribl3, merip, ribl53: text);
(* the main procedure of the program *)
const
   extrafield = 10; (* extra nfield width for the word 'coordinate' *)
var
   parameterversion: real; (* parameter version number *)
   ribl53name: string; (* name of the merged ribl *)
   ribl5from, ribl5to: integer; (* range to use from the 5' ribl *)
   ribl3from, ribl3to: integer; (* range to use from the 5' ribl *)
   matrix5, matrix3, matrix53: ribltypeptr; (* weight matricies *)
   zero: integer; (* zero coordinate on the new matrix *)
   l: integer; (* position in the new matrix *)
   lold: integer; (* position in the previous matrix *)
   b: base; (* base index for matricies *)
   spare1, spare2: string; (* spare strings for construction of names *)
   lo, hi: integer; (* low and high n(b,l) values *)
   va: integer; (* value from n(b,l) *)
   bl, bh: base; (* bases for low and high values in n(b,l) *)
   Risum: real; (* sum for averageRi *)
   nl: real; (* sum for n at a position in the ribl *)
   sum: real; (* sum for mean *)
   sumRi2: real; (* square of sum for standard deviation *)
   m: real; (* intermediate sum for mean and standard deviation *)
   variance: real; (* intermediate sum for standard deviation *)
{zzz}
begin
   writeln(output,'meri ',version:4:2);
   reset(merip);
   readln(merip, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   grabtoken(merip, ribl53name);

   readln(merip);
   readln(merip,ribl5from, ribl5to);
   readln(merip,ribl3from, ribl3to);
   readln(merip,zero);

   (* Consistency checks: order of from-to requests *)
   if ribl5from > ribl5to then begin
      writeln(output,'ribl5from (=',ribl5from:1,
                     ') must not exceed ribl5to (=',ribl5to:1,') ');
      halt;
   end;
   if ribl3from > ribl3to then begin
      writeln(output,'ribl3from (=',ribl3from:1,
                     ') must not exceed ribl3to (=',ribl3to:1,') ');
      halt;
   end;

   (* reset each file separately in case they are the same file! *)
   reset(ribl5);
   getriblmatrix(ribl5, matrix5);
   reset(ribl3);
   getriblmatrix(ribl3, matrix3);

   write(output,'Fusing ');
   writestring(output,matrix5^.riblname);
   write(output,' to ');
   writestring(output,matrix3^.riblname);
   writeln(output);

   rewrite(ribl53);
   write(output,'new name: ');
   writestring(output, ribl53name);
   writeln(output);

   (* Consistency checks: from-to requests inside actual matricies *)
   if not between(matrix5^.frombase, ribl5from, matrix5^.tobase ) then begin
      write(output,'ribl5from (=',ribl5from:1,
                     ') must be in the range of ');
      writestring(output,matrix5^.riblname);
      write(output,': from ',matrix5^.frombase:1);
      writeln(output,' to ',matrix5^.tobase:1);
      halt;
   end;
   if not between(matrix5^.frombase, ribl5to, matrix5^.tobase ) then begin
      write(output,'ribl5to (=',ribl5to:1,
                     ') must be in the range of ');
      writestring(output,matrix5^.riblname);
      write(output,': from ',matrix5^.frombase:1);
      writeln(output,' to ',matrix5^.tobase:1);
      halt;
   end;
   if not between(matrix3^.frombase, ribl3from, matrix3^.tobase ) then begin
      write(output,'ribl3from (=',ribl3from:1,
                     ') must be in the range of ');
      writestring(output,matrix3^.riblname);
      write(output,': from ',matrix3^.frombase:1);
      writeln(output,' to ',matrix3^.tobase:1);
      halt;
   end;
   if not between(matrix3^.frombase, ribl3to, matrix3^.tobase ) then begin
      write(output,'ribl3to (=',ribl3to:1,
                     ') must be in the range of ');
      writestring(output,matrix3^.riblname);
      write(output,': from ',matrix3^.frombase:1);
      writeln(output,' to ',matrix3^.tobase:1);
      halt;
   end;

   new(matrix53);
   clearstring(spare1);
   clearstring(spare2);

   (* create the header as a fusion of the two names *)
   new(matrix53^.riblheader);
   clearstring(matrix53^.riblheader^);
   attachcharacterstring('*', spare1);
   attachcharacterstring(' ', spare1);
   concatstring(spare1, matrix5^.riblname, spare2);
   attachcharacterstring('_', spare2);
   concatstring(spare2, matrix3^.riblname, matrix53^.riblheader^);
   writeln(output, 'Header: ');
   writestring(output, matrix53^.riblheader^);writeln(output);

   if ribl53name.length > 0
   then begin
      if ribl53name.letters[1] <> '+'
      then copystring(ribl53name,matrix53^.riblname)
      else copystring(matrix53^.riblheader^, matrix53^.riblname);
   end
   else copystring(matrix53^.riblheader^, matrix53^.riblname);

   (* transfer the numbers to the new matrix; display coordinates *)
   writeln(output);
{original:
}
   matrix53^.frombase := -zero + 1;
{trial:   matrix53^.frombase := -zero; }
   matrix53^.tobase   :=  matrix53^.frombase
                         + ((ribl5to-ribl5from)+1)
{trial:                         + ((ribl3to-ribl3from)-1); }
{original:
}
                         + (ribl3to-ribl3from+1) - 1;
   writeln(output,'new range: ',matrix53^.frombase:1,
                  ' to ', matrix53^.tobase:1);
   writeln(output,
                  '5''_coordinate':nfield+extrafield,
                  '3''_coordinate':nfield+extrafield,
                  'new_coordinate':nfield+extrafield);
   for l := matrix53^.frombase to matrix53^.tobase do begin
      for b := a to t do begin
         lold := (l - matrix53^.frombase) + ribl5from;
         if lold <=  ribl5to then begin
            if b = a then begin
               write(output, lold:nfield+extrafield);
               write(output, ' ':nfield+extrafield);
            end;
            (* take from the 5' matrix *)
            matrix53^.data[b,l]    := matrix5^.data[b, lold];
            matrix53^.numbers[b,l] := matrix5^.numbers[b, lold];
         end
         else begin
            (* take from the 3' matrix *)
            lold := (l - matrix53^.frombase) + ribl3from + ribl5from;
            if b = a then begin
               write(output, ' ':nfield+extrafield);
               write(output, lold:nfield+extrafield);
            end;
            matrix53^.data[b,l]    := matrix3^.data[b,lold];
            matrix53^.numbers[b,l] := matrix3^.numbers[b,lold];
         end;
      end;
      writeln(output,l:nfield+extrafield);
   end;

   (* COMPUTE remaining components of the ribl *)

   (* mean and standard deviation *)
   writeln(output);
   matrix53^.mean := 0.0;
   matrix53^.stdev := 0.0;
   for l := matrix53^.frombase to matrix53^.tobase do begin
      nl := 0;
      for b := a to t do begin
         nl := nl + matrix53^.numbers[b,l]
      end;
      sum := 0;
      sumRi2 := 0;
{zzz}
      for b := a to t do begin
         sum := sum + matrix53^.data[b,l] * (matrix53^.numbers[b,l]/nl)
      end;
      matrix53^.mean := matrix53^.mean + sum;
   end;
   writeln(output,'Mean is computed; standard deviation is set to zero.');


{zzz}
(********************) {TESTING}
{
   (* code taken from programp ri.p: *)
   (* precompute this because the difference can be so close to
      zero that negative values are obtained.  Then the square root
      function objects *)
   basestdev := sumRi2/n - (mean*mean);
   if basestdev > 0
   then stdev := sqrt(n/(n-1))*sqrt(sumRi2/n - (mean*mean))
   else stdev := 0.0;
}
{
(* Test that the mean is computed correctly by computing the 5' ribl
mean.  This worked. standard deviation is not right.*)
   (* mean *)
   matrix5^.mean := 0.0;
   variance := 0.0;
   for l := matrix5^.frombase to matrix5^.tobase do begin
      nl := 0;
      for b := a to t do begin
         nl := nl + matrix5^.numbers[b,l]
      end;
      sum := 0;
      sumRi2 := 0;
      for b := a to t do begin
         m := matrix5^.data[b,l] * (matrix5^.numbers[b,l]/nl);
         sum := sum + m;
         sumRi2 := sumRi2 + m*m;
      end;
      matrix5^.mean := matrix5^.mean + sum;
   end;
   matrix5^.stdev := sqrt(nl/(nl-1))*sqrt(sumRi2/nl
                     - (matrix5^.mean*matrix5^.mean));
writeln(output,'matrix5^.mean =', matrix5^.mean:10:6);
writeln(output,'matrix5^.stdev =', matrix5^.stdev:10:6);
halt;
}
(********************)

   (* compute consensus and anticonsensus *)
   writeln(output);
   writeln(output,'coordinate':nfield+3,
                  ' anticonsensus':nfield+6,' consensus':nfield+2);
   matrix53^.consensus := 0;
   matrix53^.anticonsensus := 0;
   for l := matrix53^.frombase to matrix53^.tobase do begin
      write(output,l:nfield+3);
      lo := +maxint;
      hi := -maxint;
      for b := a to t do begin
          if matrix53^.numbers[b,l] < lo then begin
             lo := matrix53^.numbers[b,l];
             va := lo;
             bl := b;
          end;
          if matrix53^.numbers[b,l] > hi then begin
             hi := matrix53^.numbers[b,l];
             va := hi;
             bh := b;
          end;
      end;
      matrix53^.consensus := matrix53^.consensus + matrix53^.data[bh,l];
      matrix53^.anticonsensus := matrix53^.anticonsensus + matrix53^.data[bl,l];
      writeln(output, basetochar(bl):nfield+6, basetochar(bh):nfield+2);
   end;

   (* average Ri for random sequence *)
   matrix53^.averageRi := 0.0;
   for l := matrix53^.frombase to matrix53^.tobase do begin
      Risum := 0;
      for b := a to t do begin
         Risum := Risum + matrix53^.data[b,l]
      end;
      matrix53^.averageRi := matrix53^.averageRi + Risum/4;
   end;

   (* take n to be the max of the two input ribls *)
   writeln(output);
   writeln(output, 'n for 5'' ribl: ',matrix5^.n:nfield);
   writeln(output, 'n for 3'' ribl: ',matrix3^.n:nfield);
   if matrix5^.n > matrix3^.n 
   then begin
      writeln(output, 'Using 5'' ribl n');
      matrix53^.n := matrix5^.n;
   end
   else begin
      writeln(output, 'Using 3'' ribl n');
      matrix53^.n := matrix3^.n;
   end;

   (* it makes no sense for there to be symmetry in a fused ribl! *)
   matrix53^.symmetry := 'a';
   (* clear other variables *)
   matrix53^.waves := nil;
   matrix53^.cmperbase:=0.0; matrix53^.cmperbit := 0.0;

   writeln(output);
   writeln(output,'Using 5'' ribl for Ribound, Zbound, Pbound');
   matrix53^.Ribound := matrix5^.Ribound;
   matrix53^.Zbound  := matrix5^.Zbound;
   matrix53^.Pbound  := matrix5^.Pbound;

   writeln(output);
   writeln(output,'Output resulting fused matrix, ribl53');
   putriblmatrix(ribl53, matrix53);


end;
(* end module meri.themain *)

begin
   themain(ribl5, ribl3, merip, ribl53);
1: end.

{
(* begin module scan.type *)
   (* define the ribl matrix.  Source library is prgmod.p
      REQUIRES ribl.const module
      REQUIRES base.type or book.type module
      REQUIRES wave.type module for wave definition! *)
   rblarray = array[a..t, minribl..maxribl] of real; (* real(B,L) *)
   ribltype = record
      riblname: string; (* name of the weight matrix *)
      riblheader: stringptr; (* misc header information, starting with '*' *)
      data: rblarray; (* real(B,L) *)
      numbers: array[a..t, minribl..maxribl] of integer; (* n(b,l) *)
      frombase, tobase: integer; (* range of the matrix *)
      mean, stdev: real; (* mean and standard deviation of distribution *)
      consensus: real; (* Ri value of consensus sequence *)
      anticonsensus: real; (* Ri value of anticonsensus sequence *)
      averageRi: real; (* average Ri for random sequence *)
      n: integer; (* number of sequences used to create matrix *)
      symmetry: char; (* the symmetry of the Riblmatrix *)
      waves: waveptr; (* cosine wave definitions for the matrix *)
      cmperbase, cmperbit: real; (* required for wave dash definition *)
      Ribound: real; (* the Ri boundary for this definition *) 
      Zbound: real; (* the Z boundary for this definition *) 
      Pbound: real; (* the probability boundary for this definition *) 
   end;
   ribltypeptr = ^ribltype; (* pointer to ribltype so it does
      not need to be created unless actually used *)
(* end module scan.type version = 5.46; (@ of prgmod.p 2011 Jan 20 *)
}
