program patlst (pattern,patout,patlstp,output);
(* reformats a pattern output (from the patlrn program) to a horizontal
   matrix whose width is specified by the user.
   by gary stormo, copyright 1989
modules needed: delman, delmods, auxmods *)

label 1;  (* end of program *)

const
(* begin module version *)
version = 1.09;  (* of patlst 1994 sep 5
origin aug 14, 1982 *)
(* end module version *)

(* begin module describe.patlst *)
(*
name
      patlst:  lister of patlrn output.

synopsis
      patlst(pattern: in, patout: out, patlstp: in, output: out)

files
      pattern: the input pattern matrix to be reformatted;  this is the
               output of the program patlrn.
      patout:  the output reformatted pattern matrix.
      patlstp: a parameter file for specifying the pagewidth of the patout
               file;  must contain an integer as the first thing on the first
               line, which specifies the number of matrix elements to be printed

               across a page;  if this file is empty the pagewidth is set to the
               constant 'defpagewidth'.
      output:  for messages to the user.

description
      patlst takes the output from the patlrn program and reformats the
      pattern matrix to run horizontally across the page. it is broken
      so that it fits neatly on the page.  this is useful for making
      publishable copies of the pattern matrices.

see also
      patlrn.p, patana.p

author
      gary d. stormo

bugs
      none known
*)
(* end module describe.patlst *)

      numfield = 4; (* field size for printing pattern elements *)
      defpagewidth = 30;  (* the default page width *)

(* begin module matrix.const *)
      maxmatrix = 120; (* the maximum size of a pattern matrix *)
(* end module matrix.const version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

type  base = (a,c,g,t); (* this is declared so we do not need book.type *)
(* begin module matrix.type *)
      matrix = array[base,1..maxmatrix] of integer; (* a pattern matrix *)
(* end module matrix.type version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

var   pattern,  (* the input pattern file *)
      patout,  (* the output, reformatted pattern file *)
      patlstp: text; (* the parameter file for setting pagewidth *)
      wmatrix: matrix; (* the pattern matrix *)
      pagewidth,  (* number of matrix elements printed on a page *)
      funcmin,  (* minimum value of functional sequences by this pattern *)
      nfuncmax, (* maximum value of nonfunctional sequences by this pattern *)
      beginning,  (* of the pattern matrix *)
      width: integer;  (* of the matrix *)

(* begin module package.primitive *)
(* ************************************************************************ *)
(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.57 88 mar 1 tds/gds' *)

(* begin module unlimitln *)
procedure unlimitln(var afile: text);
(* this procedure removes a stupid system dependent limit on the number of
lines that one can write to a file.  you may remove it from the code if
your system does not want or need this.  suggested method: place comments
around the contents of the procedure. *)
begin
(*      linelimit(afile, maxint); (@ set 'infinite' lines allowed for afile *)
end;
(* end module unlimitln version = 'delmod 6.57 88 mar 1 tds/gds' *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 'delmod 6.57 88 mar 1 tds/gds' *)

(* begin module copylines *)
function copylines(var fin, fout: text; n: integer): integer;
(* copy n lines of file fin to file fout.
the actual number of lines copied is returned. *)
var
      index: integer; (* the current line number *)
begin (* copylines *)
      index := 0;
      while (not eof(fin)) and (index < n) do begin
         copyaline(fin, fout);
         index := succ(index)
      end;

      copylines := index
end; (* copylines *)
(* end module copylines version = 'delmod 6.57 88 mar 1 tds/gds' *)

(* ************************************************************************ *)
(* end module package.primitive version = 'delmod 6.57 88 mar 1 tds/gds' *)

(* begin module basetochar *)
function basetochar(ba:base):char;
(* convert a base into a character *)
begin
      case ba of
         a: basetochar:='a';
         c: basetochar:='c';
         g: basetochar:='g';
         t: basetochar:='t';
      end
end;
(* end module basetochar version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

(* begin module matrix.readmatrix *)
procedure findlearnend(var pattern: text);
(* locate the end of the cyclic learning part of the pattern file *)
const
      learnmax = 16; (* the size of learnend *)
var
      (* this array contains the pattern that defines the end *)
      learnend: packed array[1..learnmax] of char;
      state: integer; (* how close we are to finding learnend *)
begin (* findlearnend *)
      reset(pattern);
      learnend := 'end of learning.';
      state := 1;
      while not eof(pattern) and (state < learnmax) do begin
         if learnend[state] = pattern^
         then state := succ(state)
         else state :=1;
         if eoln(pattern) then readln(pattern)
                          else get(pattern)
      end;
      if eof(pattern) then begin
         write  (output,'pattern matrix does not contain "');
         writeln(output,learnend,'" signal');
         halt
      end
end; (* findlearnend *)
procedure getcolon(var f: text);
(* move to the next colon in file f *)
begin (* getcolon *)
      while f^<>':' do begin
         get(f);
         if eof(f) then begin
            writeln(output,'pattern is missing colons');
            halt
         end
      end;
      get(f); (* move past the colon *)
end; (* getcolon *)

procedure readmatrix(var thefile: text; var wmatrix: matrix;
                     var beginning, width: integer);
(* read a pattern matrix wmatrix from the file thefile, returning
the aligned position for the first position (beginning) and the width *)
var   i : integer;  (* an index *)
      ba : base;  (* an index *)
begin (* readmatrix *)
      findlearnend(thefile);

      (* skip min and max values *)
      getcolon(thefile); getcolon(thefile);

      getcolon(thefile); (* skip to range *)
      read(thefile,beginning); (* read first position of the matrix *)

      getcolon(thefile); (* skip to width *)
      read(thefile,width); (* read width of the matrix *)

      if (width > maxmatrix) then begin
         writeln(output,'input matrix too large');
         halt;
      end;

      for i := 1 to width do   (* read in matrix *)
      begin
         getcolon(thefile); (* skip position label *)
         for ba := a to t do read(thefile,wmatrix[ba,i]);
         readln(thefile);
      end;
            (* set all the unused elements of wmatrix to 0 *)
      for i := (width + 1) to maxmatrix do
         for ba := a to t do wmatrix[ba,i] := 0;
end; (* readmatrix *)
(* end module matrix.readmatrix version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

procedure header;
(* write the header to the patout file *)
begin
      reset(pattern);
      rewrite(patout);
      writeln(patout,' patlst version ',version:4:2,
            '; listing of pattern matrix from:');
      readln(pattern);
      copyaline(pattern,patout);
      writeln(patout);
end;

procedure setparam;
(* set the parameter for the pagewidth *)
begin
      reset(patlstp);
      if eof(patlstp) then pagewidth := defpagewidth
      else readln(patlstp,pagewidth);
end;

function minimum(a,b: integer): integer;
(* this returns the minimum of a and b *)
begin
      if (a < b) then minimum := a
      else minimum := b;
end;

(* begin module matrix.readthresholds *)
procedure readthresholds(var pattern: text; var funcmin, nfuncmax: integer);
(* find the funcmin and nfuncmax of the pattern *)
begin (* readthresholds *)
      reset(pattern);
      findlearnend(pattern);

      getcolon(pattern);
      readln(pattern,funcmin);
      getcolon(pattern);
      readln(pattern,nfuncmax)
end; (* readthresholds *)
(* end module matrix.readthresholds version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

procedure writepattern;
(* write the pattern matrix to the file patout *)
var pageend, (* the last row of elements to be printed on this page *)
    i,j : integer;  (* indices *)
      ba: base;
begin
      write(patout,' minimum value for    functional sequences: ');
      writeln(patout,funcmin:4);
      write(patout,' maximum value for nonfunctional sequences: ');
      writeln(patout,nfuncmax:4);
      writeln(patout);
      writeln(patout,' range: ',beginning:1,' to ',
            (beginning + width -1):1,'  width: ',width:1);
      writeln(patout);
      i:= 0;
      while (i < width) do begin
         pageend := minimum(i + pagewidth,width);
               (* write position line *)
         write(patout,' position :');
         for j := i+1 to pageend do
            write(patout,(beginning+j-1):numfield);
         writeln(patout);
               (* write upper matrix border *)
         write(patout,' ':10);
         for j := 1 to (pageend -i)*numfield +1 do
            write(patout,'-');
         if (pageend = width) then writeln(patout,'-')
         else writeln(patout);
               (* write row labels and elements *)
         for ba  := a to t do begin
            write(patout,basetochar(ba):8,'  :');
            for j := i+1 to pageend do write(patout,wmatrix[ba,j]:numfield);
            if (pageend = width) then writeln(patout,':')
            else writeln(patout);
         end;
               (* write lower matrix border *)
         write(patout,' ':10);
         for j := 1 to (pageend -i)*numfield +1 do
            write(patout,'-');
         if (pageend = width) then writeln(patout,'-')
         else writeln(patout);
         writeln(patout);
         i := i + pagewidth;
      end;
end;

begin (* patlst *)
      writeln(output,'patlst ',version:4:2);
      header;
      setparam;
      readmatrix(pattern,wmatrix,beginning,width);
      readthresholds(pattern, funcmin, nfuncmax);
      writepattern;
1: end.
