program sqz(fin,fout,sqzp,output);
(* squeeze the input file to fit into fewer characters per line
   tom schneider *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.14; (* of sqz.p 1994 sep 5
origin 1986 dec 2 *)
(* end module version *)

(* begin module describe.sqz *)
(*
name
   sqz: squeeze the input file to fit into fewer characters per line

synopsis
   sqz(fin: in, fout: out, output: out, sqzp: in);

files
   fin: a text file with lines longer than 80 characters
   fout: the squeezed file.  all lines that end with the endofline symbol
      are to be continued on the next line.
      the endofline symbol is written out as the first character of the
      file, so that the unsqz program can use it.
      if the endofline symbol is found anywhere in the fin file,
      then the fout will be emptied, and the program will halt.
   sqzp: if not empty, then the first character redefines
      the endofline symbol.
   output: messages to the user

describe
   for transportation, this program allows a file to be compressed
   to fewer than 80 characters per line.

see also
   unsqz.p

author
   thomas dana schneider

bugs
   none known

technical note:
   the default endofline character is defined by a global constant
*)
(* end module describe.sqz *)

(* more constants *)
   defaultendofline = '#';  (* the character that indicates ends of lines *)
   breakpoint = 80; (* the maximum width of the output file *)

var
   fin, fout, sqzp: text; (* the input, parameter and output files
                             of the program *)

(* begin module halt *)
procedure halt;
begin
   writeln(output,'program halt');
   goto 1
end;
(* end module halt *)

(* begin module themain *)
procedure themain(var fin, fout: text);
(* the main procedure of the program *)
var
   count: integer; (* count of number of characters output so far
                      on this line *)
   endofline: char; (* the end of line character *)
begin
   writeln(output,'sqz ',version:4:2);
   rewrite(fout);
   reset(fin);
   reset(sqzp);
   if not eof(sqzp) then read(sqzp,endofline)
                    else endofline := defaultendofline;
   write(fout,endofline); (* tell the unsqz program what to use *)
   writeln(fout,' squeezed file from sqz ',version:4:2);

   while not eof(fin) do begin
      count := 0;
      while not eoln(fin) do begin
         if fin^=endofline then begin
            writeln(output,'the character ',endofline,
                           ' must not appear in the file');
            writeln(output,'the fout file is being destroyed.');
            rewrite(fout);
            halt
         end;
         if count=breakpoint-1 then begin
            writeln(fout,endofline);
            count := 0
         end;
         write(fout,fin^);
         get(fin);
         count := count + 1;
      end;
      readln(fin);
      writeln(fout)
   end;
end;
(* end module themain *)

begin
   themain(fin,fout);
1: end.
