program swt(swtp, list, output);
(* swt: switch test for I information computation

  Tom Schneider
  permanent email: toms@alum.mit.edu
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms/


 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.00; (* of swt.p 1998 mar 16
origin 1998 mar 16 *)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.swt *)
(*
name
   swt: switch test for I information computation

synopsis
   swt(swtp: in, list: out, output: out)

files
   list:  multiple line detailed description of file 2, etc

   swtp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

   output: messages to the user

description

   Test whether Iab+Ibc = Iac+Icb.  The fact that they differ shows that this
method of computing "information" is sensitive to the order of tests, even
though the final results are the same.

examples

swt 1.00
 pa =    0.50000
 pb =    0.40000
 pc =    0.30000

For a->b->c we have: 
Iab = sum pb log pb/pa
Ibc = sum pc log pc/pb
iab + ibc =    0.06021 "bits"

For a->c->b we have:
Iac = sum pc log pc/pa
Icb = sum pb log pb/pc
iac + icb =    0.15129 "bits"

documentation

@article{Schneider.edmm,
author = "T. D. Schneider",
title = "Theory of Molecular Machines.
{II. Energy} Dissipation from Molecular Machines",
journal = "J. Theor. Biol.",
volume = "148",
thenumber = "1",
pages = "125-137",
note = "http://www-lecb.ncifcrf.gov/$\sim$toms/paper/edmm/",
year = 1991}

see also

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.swt *)

var
   swtp, (* file used by this program *)
   list: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

function doone(pa,pb: real): real;
(* compute "information" I for probabilities pa and pb *)
var
  iab: real; (* information of a -> b *)
begin
  iab := pb    *ln(   pb /   pa )/ln(2) +
         (1-pb)*ln((1-pb)/(1-pa))/ln(2);
  doone := iab
end;

(* begin module swt.themain *)
procedure themain(var swtp, list: text);
(* the main procedure of the program *)
var
   parameterversion: real; (* parameter version number *)
   pa, pb, pc: real; (* input probabilities *)
   iab, ibc, iac, icb: real; (* output "information" *)
begin
   writeln(output,'swt ',version:4:2);
   reset(swtp);
   readln(swtp, parameterversion);
   if parameterversion < updateversion then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   readln(swtp, pa, pb, pc);

   rewrite(list);
   writeln(list,'swt ',version:4:2);
   writeln(list,' pa = ',pa:10:5);
   writeln(list,' pb = ',pb:10:5);
   writeln(list,' pc = ',pc:10:5);

   writeln(list);
   writeln(list,'For a->b->c we have: ');
   writeln(list, 'Iab = sum pb log pb/pa');
   writeln(list, 'Ibc = sum pc log pc/pb');

   iab := doone(pa,pb);
   ibc := doone(pb,pc);

   writeln(list, 'iab + ibc = ', (iab+ibc):10:5, ' "bits"');
 
   writeln(list);
   writeln(list,'For a->c->b we have:');
   writeln(list, 'Iac = sum pc log pc/pa');
   writeln(list, 'Icb = sum pb log pb/pc');

   iac := doone(pa,pc);
   icb := doone(pc,pb);
   writeln(list, 'iac + icb = ', (iac+icb):10:5, ' "bits"');

end;
(* end module swt.themain *)

begin
   themain(swtp, list);
1: end.
