program untitle(input,output);
(* untitle: remove titles from bbl file
   by thomas schneider, copyright (c) 1988  
module libraries required: delman, prgmods *)
  
label 1; (* end of program *) 
  
const 
(* begin module version *)
version = 1.17; (* of untitle, 1988 july 2
origin 1988 feb 25 from domods *)
(* end module version *)
  
(* begin module describe.untitle *) 
(*
name
   untitle: remove titles from bbl file
  
synopsis
   untitle(input: in, output: out) 
  
files 
   input: a bbl file from bibtex.
   output: a bbl file without titles.

description 
   Titles are removed by deleting between the two copies of the
   '\newblock' strings, leaving the second one.  If the first '\newblock'
   contains the italics indicator, '{\it', then the program realizes
   that this must be a book title, and it keeps the title.

author
   Thomas D. Schneider 
  
*)
(* end module describe.untitle *)

(* begin module interact.const *) 
      maxstring = 150; (* the maximum string *) 
(* end module interact.const version = 'prgmod 3.97  85 may 5 tds'; *)
  
(* begin module untitle.filler.const *) 
      fillermax = 10; (* the size of the filler array for a string *) 
(* end module untitle.filler.const *)
  
type  

(* begin module interact.type *)
      string = record (* a string of characters *)  
         letters: array[1..maxstring] of char; (* the letters in the string *)
         length: integer; (* the number of characters in the string *)
         current: integer; (* the letter we are working on *) 
      end;  
(* end module interact.type version = 'prgmod 3.97  85 may 5 tds'; *)

(* begin module trigger.type *) 
      trigger = record (* an object to be searched for *) 
         seek: string; (* the characters looked for *)  
         state: integer; (* how close to triggering we are *) 
         skip: boolean; (* trigger not found- skip the line *)
         found: boolean (* the trigger was found *) 
      end;  
(* end module trigger.type version = 'prgmod 3.97  85 may 5 tds'; *)

(* begin module filler.type *)
      (* the following is an array used to fill a string. 
      it is convenient to have it much shorter than the maxstring, so that  
      it is easy to fill the string using procedure fillstring. 
      the user must declare the value of constant fillermax. *) 
      filler = packed array[1..fillermax] of char;
(* end module filler.type version = 'prgmod 3.97  85 may 5 tds'; *)

procedure halt; 
(* stop the program.  the procedure performs a goto to the end of the   
   program.  you must have a label: 
      label 1;
   declared, and also the end of the program must have this label:  
      1: end. 
   examples are in the module libraries.
   this is the only goto in the delila system. *) 
begin 
      writeln(output,' program halt.'); 
      goto 1
end;  
(* end module halt version = 'prgmod 3.97  85 may 5 tds'; *)
  
(* begin module interact.clearstring *) 
procedure clearstring(var ribbon: string);  
(* empty the string *)
var   index: integer; (* to the ribbon *) 
begin (* clearstring *) 
      with ribbon do begin  
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0; 
         current := 0;
      end 
end; (* clearstring *)
(* end module interact.clearstring version = 'prgmod 3.97  85 may 5 tds'; *)

(* begin module interact.writestring *) 
procedure writestring(var tofile: text; var s: string); 
(* write the string s to file tofile, no writeln *) 
var   i: integer; (* index to s *)  
begin (* writestring *) 
      with s do for i := 1 to length do write(tofile, letters[i]) 
end; (* writestring *)
(* end module interact.writestring version = 'prgmod 3.97  85 may 5 tds'; *)

(* begin module trigger.proc *) 
(* this module allows one to scan a series of characters, as from 
an array or a file, and to "trigger" or detect a simple string  
in the series.  the advantage of the trigger is that several triggers 
can "observe" a stream of characters at once, each looking for a
different thing.  
some other modules required: interact.const, interact.type *) 
  
procedure resettrigger(var t: trigger); 
(* reset the trigger to ground state *) 
begin (* resettrigger *)
      with t do begin 
         state := 0;
         skip := false; 
         found := false 
      end 
end; (* resettrigger *) 
  
procedure testfortrigger(ch: char; var t: trigger); 
(* look at the character ch.
   if it is part of the trigger (at the current trigger state), 
       then the trigger state goes higher.  
   if it is not part of the trigger then the trigger state is reset,  
      skip is true and one should skip onward to find the trigger.  
   if the trigger is found, found is true. *) 
begin (* testfortrigger *)
      with t do begin 
         state := succ(state);  
(*         if debugging then begin  
            writestring(list,seek); 
            writeln(list,'testfortrigger seek.letters[',state:1,']:', 
                           seek.letters[state],' ch:',ch);
         end;*) 
         if seek.letters[state] = ch  
         then begin 
            skip := false;
            if state = seek.length then found := true 
                                   else found := false  
         end
         else begin (* reset trigger *) 
            state := 0; 
            skip := true; 
            found := false
         end
      end 
end; (* testfortrigger *) 
(* end module trigger.proc version = 'prgmod 3.97  85 may 5 tds'; *)

(* begin module filler.fillstring *)
procedure fillstring(var s: string; a: filler); 
(* this procedure makes it reasonably easy to fill the string s with
characters.  one calls the procedure as: *) 
(*                           1         2         3         4         5 *) 
(*                  12345678901234567890123456789012345678901234567890 *) 
(*   fillstring(s, 'this-is-the-string                                ');   
the two comments make it easy to line the characters up. also, for this 
example, it was assumed that the length of filler as defined by the 
constant fillermax was 50. *) 
var 
      length: integer; (* of the string without trailing blanks *)  
      index: integer; (* of s *)  
begin (* fillstring *)  
      clearstring(s); 
      length := fillermax;  
      while (length > 1) and (a[length] = ' ') do length := pred(length); 
      if (length = 1) and (a[length] = ' ') then begin  
         writeln(output, 'fillstring: the string is empty');
         halt 
      end;  
  
      for index := 1 to length do s.letters[index] := a[index]; 
      s.length := length; 
      s.current := 1
end; (* fillstring *) 
(* end module filler.fillstring version = 'prgmod 3.97  85 may 5 tds'; *)

(* begin module filler.filltrigger *) 
procedure filltrigger(var t: trigger; 
                      a: filler); 
(* fill the trigger t *)  
begin (* filltrigger *) 
         fillstring(t.seek,a) 
end; (* fillstring *) 
(* end module filler.filltrigger version = 'prgmod 3.97  85 may 5 tds'; *)

(* ********************************************************************** *)
(* ********************************************************************** *)
(* ********************************************************************** *)

(* begin module untitle.themain *)
procedure themain(var fin, fout: text);  
(* the main procedure of the program *)
var 
   bibitem: trigger; (* start of a bibliography item *)
   it: trigger; (* indication that the title is from a book *)
   ch: char; (* a character read from fin *)  
   copying: boolean; (* true if we copy from fin to fout *)
   nbc: integer; (* newblock count.  the first is the title, the
      others are the reference *)
   newblock: trigger; (* a trigger for the newblock *) 
begin 
(*                            1 *)
(*                   1234567890 *)
filltrigger(newblock,'\newblock '); (* first newblock is title *)
filltrigger(bibitem, '\bibitem  '); (* start of a citation's data *)
filltrigger(it,      '{\it      '); (* book titles are in italics *)
   copying := true;

   while not eof(fin) do begin 
      resettrigger(bibitem);
      resettrigger(newblock);
      resettrigger(it);

      while not eoln(fin) do begin
         read(fin,ch);
         if copying then write(fout, ch);

         testfortrigger(ch,bibitem); 

         if bibitem.found
         then begin (* begin processing for a bib item *)
            copying := true;
            nbc := 0;
         end
         else begin
            testfortrigger(ch,newblock); 
            if newblock.found then begin
               nbc := nbc + 1;
               if nbc = 1
               then copying := false  (* normally don't copy first newblock *)
               else copying := true;  (* copy everything else *)
            end;

            testfortrigger(ch,it);
            if it.found and (not copying) then begin
               (* book title, DO copy it *)
               (* restore the missing parts *)
               write(fout,' ');
               writestring(fout,it.seek);
               copying := true;
            end;
         end;
    
      end;
      readln(fin);
      if copying then writeln(fout);
   end
end;  
(* end module untitle.themain *)
 
begin
   themain(input, output);
1: end.
