program whatch(fin, fout, output);
(* whatch: what characters are in a file?
   Thomas Schneider
   copyright (c) 1985

   module libraries: delman, prgmods *)

const
(* begin module version *)
version = 1.13; (* of whatch 1997 February 26
1996 feb 13 spawning of wtch
1985 apr 20
origin 1982 sep 5 *)
(* end module version *)

(* begin module describe.whatch *)
(*
name
      whatch: what characters are in a file?

synopsis
      whatch(fin: in, fout: out, output: out)

files
      fin: the file to be studied

      fout: an alphabetic list of the characters in the file, giving:
         the character,
         the ordinal number of the character (pascal ord function),
         how many such characters are in the file,
         and the percent of the character in the file.

      output: messages to the user

description
      sometimes it is necessary to determine what characters are in a file.
      if the file is very large, it is not possible to do this by hand.

see also
      wtch.p

author
      thomas schneider

bugs
      none known

technical notes
      the constant maxchars determines the number of characters accepted.

*)
(* end module describe.whatch *)


(* more constants *)
      maxchars = 255; (* the last character acceptable *)

type
      charident = record (* the character and how many found *)
         ch: char;
         nu: integer
      end;

var
      fin, (* input file *)
      fout: (* the list of characters *)
         text;

      chars: array[0..maxchars] of charident; (* information on the chars *)
      index: 0..maxchars; (* an index for chars *)
      count: integer; (* of all the characters *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 'prgmod 3.96  85 mar 18 tds'; *)

begin (* whatch *)
      (* initialize *)
      writeln(output,' whatch ',version:4:2);
      reset(fin);
      rewrite(fout);
      writeln(fout,' whatch ',version:4:2);
      write(fout,' ');
      if not eof(fin) then copyaline(fin,fout);
      reset(fin);
      writeln(fout);

      (* clear the chars array *)
      for index := 0 to maxchars do with chars[index] do begin
         ch := ' ';
         nu := 0
      end;

      (* scan the fin *)
      count := 0;
      while not eof(fin) do begin
         count := succ(count);
         if ord(fin^) > maxchars then begin
            writeln(fout,' character with ord = ',ord(fin^):1,
                         ' this is greater than maxchars (',maxchars:1,')');
            writeln(output,' increase maxchars')
         end
         else if ord(fin^) < 0 then begin (* rare but interesting... *)
            writeln(fout,' character with negative ord: ',ord(fin^):1)
         end
         else begin
            with chars[ord(fin^)] do begin
               if nu > 0
               then nu := succ(nu)
               else begin
                  nu := 1;
                  ch := fin^
               end
            end
         end;

         get(fin)
      end;

      (* put out report to fout *)
      writeln(fout);
      writeln(fout,' ',count:1,' characters found');
      if count <> 0 then begin
         writeln(fout);
         writeln(fout,' char ord(char) count percent');
         for index := 0 to maxchars do begin
            if chars[index].nu <> 0 then begin
               write(fout,'  ',chars[index].ch);
               write(fout,'     ',index:5);
               write(fout,'   ',chars[index].nu:5);
               write(fout,' ',100*(chars[index].nu/count):7:2);
               if chars[index].ch <> chr(index) then begin
                  write(fout,' warning: program error,');
                  write(fout,' actual character is different')
               end;
               writeln(fout)
            end
         end
      end
end. (* whatch *)
